Package: ipdDb
Title: IPD IMGT/HLA and IPD KIR database for Homo sapiens
Description: All alleles from the IPD IMGT/HLA 
  <https://www.ebi.ac.uk/ipd/imgt/hla/> and IPD KIR 
  <https://www.ebi.ac.uk/ipd/kir/> database for Homo sapiens. 
  Reference: 
    Robinson J, Maccari G, Marsh SGE, Walter L, Blokhuis J, Bimber B, Parham P, 
    De Groot NG, Bontrop RE, Guethlein LA, and Hammond JA
    KIR Nomenclature in non-human species
    Immunogenetics (2018), in preparation.
Version: 1.29.0
Author: Steffen Klasberg
Maintainer: Steffen Klasberg <klasberg@dkms-lab.de>
Depends: R (>= 3.5.0), methods, AnnotationDbi (>= 1.43.1),
        AnnotationHub
Imports: Biostrings, GenomicRanges, RSQLite, DBI, IRanges, stats,
        assertthat
License: Artistic-2.0
URL: https://github.com/DKMS-LSL/ipdDb
BugReports: https://github.com/DKMS-LSL/ipdDb/issues/new
organism: Homo sapiens
species: Homo sapiens
biocViews: GenomicVariation, SequenceMatching, VariantAnnotation,
        DataRepresentation,AnnotationHubSoftware
Suggests: knitr, rmarkdown, testthat
RoxygenNote: 6.1.0
Encoding: UTF-8
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/ipdDb
git_branch: devel
git_last_commit: 1180c38
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 22:24:56 UTC; biocbuild

# Generated by roxygen2: do not edit by hand

export(loadHlaData)
export(loadKirData)
exportMethods(columns)
exportMethods(getAlleles)
exportMethods(getClosestComplete)
exportMethods(getLoci)
exportMethods(getReference)
exportMethods(getStructure)
exportMethods(keys)
exportMethods(keytypes)
exportMethods(select)
import(AnnotationDbi)
import(AnnotationHub)
import(DBI)
import(RSQLite)
import(assertthat)
import(methods)
importFrom(Biostrings,DNAStringSet)
importFrom(GenomicRanges,GRanges)
importFrom(IRanges,IRanges)
importFrom(stats,setNames)

ipdDb
=============

Database package of HLA and KIR alleles from the 
[IPD IMGT/HLA](https://www.ebi.ac.uk/ipd/imgt/hla/) | [Github](https://github.com/ANHIG/IMGTHLA) and 
[IPD KIR](https://www.ebi.ac.uk/ipd/kir/) | [Github](https://github.com/ANHIG/IPDKIR) databases.

Reference:

>Robinson J, Maccari G, Marsh SGE, Walter L, Blokhuis J, Bimber B, Parham P, De Groot NG, Bontrop RE, Guethlein LA, and Hammond JA  
KIR Nomenclature in non-human species  
Immunogenetics (2018), in preparation



This package holds all information from the IPD databases.
For HLA this is limited to HLA-A, -B, -C, -DPB1, -DQB1 and -DRB1.

For alleles which are not known in full-length, also the closest full-length allele is stored.

The data is stored in an SQLite database and its compatible with Bioconductor's database package rules, 
i.e. the `select` method including `columns`, `keytypes` and `keys`.

Additionally, some helper functions are implemented to fetch all contained loci, all alleles of a locus,
the sequences of alleles and the sequence of closest allele which is available in full-length for an allele.

```r
library(ipdDb)

hla <- loadHlaData()
## get all loci stored in the db
available_loci <- hla$getLoci()

## get all alleles of a locus
alleles <- hla$getAlleles(available_loci[1])
alleles <- hla$getAlleles("HLA-A")

## get all sequences of a bunch of alleles as DNAStringSet
sequences <- hla$getReference(alleles)
sequences <- hla$getReference(c("HLA-A*01:01:01:01", "HLA-A*01:01:01:03" ))

## get the closest complete reference for ONE allele as DNAStringSet
closest_complete <- hla$getClosestComplete(alleles[1])
closest_complete <- hla$getClosestComplete("HLA-A*01:01:01:01")

## Get the gene structure for a bunch of alleles as GRanges object
structures <- hla$getStructure(alleles)
structures <- hla$getStructure(c("HLA-A*01:01:01:01", "HLA-A*01:01:01:03" ))

## specifiy the database version
hla <- loadHlaData(version = "3.32.0")

```

## Load database
#' Load the IPD IMGT/HLA database
#' 
#' @param version Either a valid version of the IPD IMGT/HLA database or 
#'   "Latest" to fetch the latest version
#' @return an \code{\link{IpdDb}} object containing the database.
#' @usage loadHlaData(version = "Latest")
#' @examples 
#' ## Load the HLA database
#' hla <- loadHlaData()
#' @export
loadHlaData <- function(version = "Latest") {
  accNumber <- switch(version,
                      "Latest" = "AH66146",
                      "3.32.0" = "AH63658",
                      "3.33.0" = "AH66146",
                      NULL)
  if (is.null(accNumber))
    stop(sprintf("IPD IMGT/HLA version %s in not yet included or exist at all", 
                 version))
  .getData(accNumber)
}
#' Load the IPD KIR database
#' 
#' @param version Either a valid version of the IPD KIR database or "Latest" to
#'   fetch the latest version
#' @return an \code{\link{IpdDb}} object containing the database.
#' @usage loadKirData(version = "Latest")
#' @examples
#' ## Load the KIR database
#' kir <- loadKirData()
#' @export
loadKirData <- function(version = "Latest") {
  accNumber <- switch(version,
                      "Latest" = "AH63659",
                      "2.7.1"  = "AH63659",
                      NULL)
  if (is.null(accNumber))
    stop(sprintf("IPD KIR version %s in not yet included or exist at all", 
                 version))
  .getData(accNumber)
}


### Helper ###
.getData <- function(accNumber) {
  hub <- AnnotationHub()
  hub[[accNumber]]
}

context("dbMethods")

hla <- loadHlaData()
kir <- loadKirData()
test_that("HLA db can be loaded", {
  expect_is(hla, "IpdDb")
})

test_that("HLA db has valid version", {
  version <- hla$getDbVersion()
  expect_is(version, "character")
  expect_true(grepl("\\d+\\.\\d+\\.\\d+", version))
})

test_that("KIR db can be loaded", {
  expect_is(kir, "IpdDb")
})

test_that("KIR db has valid version", {
  version <- kir$getDbVersion()
  expect_is(version, "character")
  expect_true(grepl("\\d+\\.\\d+\\.\\d+", version))
})

test_that("keytypes and keys method works", {
  kt <- keytypes(hla)
  expect_is(kt, "character")
  ## test that all keytypes are valid and return their keys
  expect_true(all(vapply(kt, function(k, hla) {
    is.character(keys(hla, k))
    }, FUN.VALUE = logical(1), hla = hla)))
  ## Test that a wrong keytype does not work
  expect_error(keys(hla, "testWrongKey"))
})

test_that("columns method works", {
  kt <- keytypes(hla)[1]
  cols <- columns(hla)
  keys <- keys(hla, keytype = kt)[1]
  expect_is(cols, "character")
  ## test that all columns are valid
  expect_true(all(tolower(cols) %in% colnames(select(hla, keys, cols, kt))))
  ## Test if a wrong column gives an error
  expect_error(select(hla, keys, c("blubb", cols), kt))
})

test_that("select method work", {
  kt <- keytypes(hla)[1]
  cols <- columns(hla)[1:2]
  keys <- keys(hla, kt)[1:5]
  res <- select(hla, keys, cols, kt)
  expect_is(res, "data.frame")
  expect_equal(length(res), length(cols))
  expect_equal(length(unique(res[[tolower(kt)]])), length(keys))
})

## define IpdDb class
##

#' The database class for storing allele data from IPD.
#' 
#' This class extends the 
#' \code{\link[AnnotationDbi:AnnDbObj-class]{AnnotationDbi::AnnDbObj-class}}
#' object by higher level methods for sequence and annotation retrieval.
#' blubb
#' 
#' @aliases IpdDb 
#' @rdname IpdDb
#' @seealso 
#'   \code{\link[AnnotationDbi:AnnDbObj-class]{AnnotationDbi::AnnDbObj-class}}
#' @slot getDbVersion() Get the version of the original ipd database
#' @slot getLoci() get all loci from a database, see 
#'   \code{\link{getLoci}}.
#' @slot getReference(alleles) Get the reference sequence for alleles, see 
#'   \code{\link{getReference}}.
#' @slot getStructure(alleles) Get the structures of alleles, see 
#'   \code{\link{getStructure}}.
#' @slot getClosestComplete(allele) Get the closest full-length reference 
#'   sequence of one allele, see \code{\link{getClosestComplete}}.
#' @slot getAlleles(locus) Get all alleles of a locus, see 
#'   \code{\link{getAlleles}}.
#' 
#' @examples 
#' ## load the data 
#' hla <- loadHlaData()
#' ## get all valid keytypes
#' kts <- keytypes(hla)
#' ## get all valid columns
#' cols <- columns(hla)
#' ## get the keys of one keytype
#' kt <- kts[1]
#' keys <- keys(hla, kt)
#' ## Get data of the two first columns for the first 10 keys 
#' cols <- cols[1:10]
#' res <- select(hla, keys, cols, kt)
#'   
.IpdDb <- setRefClass("IpdDb", contains="AnnotationDb", 
                      methods = list(
                        getDbVersion = function(.self) {
                          info <- dbInfo(.self$conn)
                          info[info$name == "DB_VERSION", "value"]
                        },
                        getLoci = function(.self) {
                          getLoci(.self)
                        },
                        getReference = function(.self, allele) {
                          getReference(.self, allele)
                        },
                        getStructure = function(.self, allele) {
                          getStructure(.self, allele)
                        },
                        getClosestComplete = function(.self, allele, 
                                                      locus = NULL) {
                          getClosestComplete(.self, allele, locus)
                        },
                        getAlleles = function(.self, locus) {
                          getAlleles(.self, locus)
                        }
                      ))


## Set generics and methods
## Select interface
#' Columns method queries all columns present in the database
#' @rdname IpdDb 
#' @aliases columns
#' @usage columns(x)
#' @export 
setMethod("columns", "IpdDb", function(x){.cols(x)})
#' The keytypes method returns all columns that can be used as keytypes to 
#'   retrieve keys.
#' @rdname IpdDb
#' @aliases keytypes
#' @usage keytypes(x)
#' @export 
setMethod("keytypes", "IpdDb", function(x){.cols(x)})
#' The keys method gets all keys for a give keytype, i.e. all unique entries of
#'   the keytype.
#' @rdname IpdDb
#' @aliases keys
#' @usage keys(x, keytype, ...)
#' @export 
setMethod("keys", "IpdDb", function(x, keytype, ...){.keys(x, keytype)})
#' The select method queries the database for the values of all give columns for
#'   a set of keys that match a specified keytype.
#' @aliases select
#' @rdname IpdDb 
#' @param x the IpdDb object
#' @param keys The keys for which columns should be selected by select()
#' @param columns The columns to retrieve by select
#' @param keytype The keytype for which the keys are retrieved
#' @param ... Additional arguments. Not used now.
#' @return character vector (keys, columns, keytypes) or a data.frame (select).
#' @usage select(x, keys, columns, keytype, ...)
#' @export 
setMethod("select", "IpdDb", function(x, keys, columns, keytype, ...){
  .select(x, keys, columns, keytype)
})

## Sequence retrieval
setGeneric("getLoci", signature = "x", function(x) {
  standardGeneric("getLoci")
})
#' Get loci
#' 
#' Get all available loci of the KIR or HLA database
#' @aliases getLoci
#' @usage getLoci(x)
#' @param x The database connection; an \code{\link{IpdDb}} object.
#' @return A vector of available loci in the database.
#' @examples
#' ## Load the database 
#' hla <- loadHlaData()
#' ## Get the loci
#' loci <- getLoci(hla)
#' 
#' @export 
setMethod("getLoci", signature = "IpdDb", function(x){
  .getLoci(x)
})

setGeneric("getReference", signature = "x", function(x, allele) {
  standardGeneric("getReference")
})

#' Get reference sequences
#' 
#' Get the reference sequences for alleles.
#' 
#' @aliases getReference
#' @usage getReference(x, allele)
#' @param x The database connection; an \code{\link{IpdDb}} object.
#' @param allele The alleles of interest as a character vector.
#' @return A \code{\link[Biostrings:XStringSet-class]{Biostrings:DNAStringSet}}
#' object with all references.
#' @examples
#' ## Load the database 
#' hla <- loadHlaData()
#' ## Get the loci
#' loci <- getLoci(hla)
#' ## Get alleles of a locus
#' alleles <- getAlleles(hla, loci[1])
#' allelesOfInterest <- alleles[1:10]
#' ## Get the sequences
#' seqs <- getReference(hla, allelesOfInterest)
#' 
#' @export 
setMethod("getReference", signature = "IpdDb", function(x, allele){
  .getReference(x, allele)
})

setGeneric("getStructure", signature = "x", function(x, allele) {
  standardGeneric("getStructure")
})
#' Get gene structures
#' 
#' Get the gene structures for alleles.
#' 
#' @aliases getStructure
#' @usage getStructure(x, allele)
#' @param x The database connection; an \code{\link{IpdDb}} object.
#' @param allele The alleles of interest as a character vector.
#' @return A \code{\link[GenomicRanges:GRanges-class]{GenomicRanges:GRanges}} 
#' object with all gene
#'   structures.
#' @examples
#' ## Load the database 
#' hla <- loadHlaData()
#' ## Get the loci
#' loci <- getLoci(hla)
#' ## Get alleles of a locus
#' alleles <- getAlleles(hla, loci[1])
#' allelesOfInterest <- alleles[1:10]
#' ## Get the structures
#' seqs <- getStructure(hla, allelesOfInterest)
#' 
#' @export 
setMethod("getStructure", signature = "IpdDb", function(x, allele){
  .getStructure(x, allele)
})

setGeneric("getClosestComplete", signature = "x", function(x, allele, 
                                                           locus = NULL) {
  standardGeneric("getClosestComplete")
})
#' Get closest full-length sequence
#' 
#' Get the sequence of the closest allele which for which a full-length 
#'   sequence is available.
#' 
#' @aliases getClosestComplete
#' @usage getClosestComplete(x, allele, locus = NULL)
#' @param x The database connection; an \code{\link{IpdDb}} object.
#' @param allele A single allele as a string.
#' @param locus optional parameter used if the allele identifier is not found.
#' @return A \code{\link[Biostrings:XStringSet-class]{Biostrings:DNAStringSet}}
#' object with the sequence 
#' of the closest full-length allele.
#' @examples
#' ## Load the database 
#' hla <- loadHlaData()
#' ## Get the loci
#' loci <- getLoci(hla)
#' ## Get alleles of a locus
#' alleles <- getAlleles(hla, loci[1])
#' alleleOfInterest <- alleles[1]
#' ## Get the closest complete sequence
#' seqs <- getClosestComplete(hla, alleleOfInterest, loci[1])
#' 
#' @export 
setMethod("getClosestComplete", signature = "IpdDb", function(x, allele, 
                                                              locus = NULL){
  .getClosestComplete(x, allele, locus)
})

setGeneric("getAlleles", signature = "x", function(x, locus) {
  standardGeneric("getAlleles")
})
#' Get alleles
#' 
#' Get all alleles of a given locus.
#' 
#' @aliases getAlleles
#' @usage getAlleles(x, locus)
#' @param x The database connection; an \code{\link{IpdDb}} object.
#' @param locus A single locus as a string.
#' @return A character vector with all alleles of the give locus.
#' @examples
#' ## Load the database 
#' hla <- loadHlaData()
#' ## Get the loci
#' loci <- getLoci(hla)
#' ## Get alleles of a locus
#' alleles <- getAlleles(hla, loci[1])
#' 
#' @export 
setMethod("getAlleles", signature = "IpdDb", function(x, locus){
  .getAlleles(x, locus)
})

context("seqMethods")
hla <- loadHlaData()

test_that("getReference works as expected", {
  locus <- getLoci(hla)[1]
  alleles <- getAlleles(hla, locus)
  ## Test single allele
  seq <- getReference(hla, alleles[1])
  expect_is(seq, "DNAStringSet")
  ## Test multiple alleles
  seq <- getReference(hla, alleles[1:10])
  expect_is(seq, "DNAStringSet")
  expect_error(getReference(hla, "xx"))
  expect_error(getReference(hla, 1))
  expect_error(getReference(hla, TRUE))
  expect_error(getReference(hla, locus))
})

test_that("getLoci works as expected", {
  loci <- getLoci(hla)
  expect_is(loci, "character")
  ## check if all loci begin with "HLA" or "KIR"
  expect_true(all(vapply(loci, function(l) startsWith(l, "HLA-") || 
                           startsWith(l, "KIR"), FUN.VALUE = logical(1))))
})

test_that(".matchAllele finds correct alleles", {
  locus <- "HLA-A"
  allAlleles <- getAlleles(hla, locus)
  targetAllele <- "HLA-A*01:01:01:01"
  
  ## test partial alleles 
  allele <- "01:01:01:01"
  expect_identical(.matchAllele(hla, allele, locus), targetAllele)
  
  ## test truncated alleles; shorter resolution
  allele <- "HLA-A*01:01:01"
  targetAlleles <- allAlleles[startsWith(allAlleles, allele)]
  expect_identical(.matchAllele(hla, allele, locus), targetAlleles)
  
  allele <- "HLA-A*9999"
  expect_identical(.matchAllele(hla, allele, locus), allAlleles)
  
  expect_error(.matchAllele(hla, 123, locus))
  expect_error(.matchAllele(hla, TRUE, locus))
  expect_error(.matchAllele(hla, allele, "testlocus"))
})

test_that("getAlleles works correct", {
  locus <- "HLA-A"
  ## Get the target alleles using plain methods
  keys <- keys(hla, keytype = "locus" )
  targetAlleles <- select(hla, keys = locus, 
                          columns = "allele_name", 
                          keytype = "locus")$allele_name
  expect_identical(getAlleles(hla, locus), targetAlleles)
  expect_error(getAlleles(hla, "invalidLocus"))
})

test_that("getStructure works correct",{
  alleles <- getAlleles(hla, getLoci(hla)[1])
  expect_is(getStructure(hla, alleles[1]), "GRanges")
  expect_is(getStructure(hla, alleles[1:10]), "GRanges")
  
  ## Test for invalid keys
  expect_error(getStructure(hla, "test"))
  expect_error(getStructure(hla, 1))
  expect_error(getStructure(hla, TRUE))
})

test_that("closestComplete gives correct sequences", {
  alleles <- getAlleles(hla, getLoci(hla)[1])
  expect_is(getClosestComplete(hla, alleles[1]), "DNAStringSet")
  expect_error(getClosestComplete(hla, alleles[1:100]))
  expect_error(getClosestComplete(hla, "test"))
  expect_error(getClosestComplete(hla, TRUE))
  expect_error(getClosestComplete(hla, 123))
})

## This script holds condensed information on how the hla and kir databases
## for the ipdDb package are created. This requires four steps:
## 1. download the data from github:
##   Hla: https://github.com/ANHIG/IMGTHLA We use the hla/hla.xml.zip in that
##     repo
##   Kir: https://github.com/ANHIG/IPDKIR We use the KIR.dat, as there is not
##    yet a reliable xml file
## 2. Read the file and store it in a data.frame. Extract the sequences as
##   DNAStringSet and the structures as GRanges objects.
## 3. Find the closest full-length allele. All alleles of one locus
##   are put into groups of "feature compositions". We have a group for each
##   kind of partial allele, e.g. all alleles were only exon 2 and 3 are known
##   are put together. All full-length alleles are also grouped and known part
##   of the "feature composition" is extracted, e.g. only exon 2 and 3.
##   A distance matrix is build from all sequences in the group. The full-length
##   allele with the minimal distance is considered the closest full-length
##   allele.
## 4. The dataframes are put into an SQLite database based on the AnnotationDbi
##   vignette.
##




library(BSgenome)
library(dplyr)
library(foreach)
library(Biostrings)
library(BiocParallel)
### Get IPD Database ####
###
###
#' Clone the IMGTHLA repo on github
#'
#' @note Due to limitations in the \pkg{git2r} this will not work
#' behind a proxy.
#' @param family one of "kir" or "hla". If none is provided it will fetch HLA.
#' @param url url to fetch from
#' @return The path to the local repo
fetchIPD <- function(family="hla", url=NULL) {
  hla_url <- "https://github.com/ANHIG/IMGTHLA"
  kir_url <- "https://github.com/ANHIG/IPDKIR"
  if (is.null(url))
    url <- ifelse(tolower(family) == "kir", kir_url, hla_url)
  assertthat::assert_that(RCurl::url.exists(url))
  assertthat::assert_that(all(
    requireNamespace("git2r", quietly = TRUE)
  ))
  tmp <- tempfile()
  # Currently not working behind proxy
  git2r::clone(url, tmp)
  tmp
}

#' Fetch the IMGT/HLA hla.xml file
#'
#' @param dbpath the path to the xml file
#' @return An object of class (S3) \code{XMLInternalDocument}.
readHlaXml <- function(dbpath) {
  dbfile <- normalizePath(file.path(dbpath, "xml", "hla.xml.zip"),
                          mustWork = TRUE)
  tfile <- unzip(zipfile = dbfile, exdir = dbpath)
  tfile <- tfile[grepl(pattern = "^hla.xml", x = basename(tfile))]
  doc <- xml2::read_xml(tfile)
  unlink(dbpath, force = TRUE)
  doc
}

#' Fetch the KIR.da file in EMBL format ##
#'
#' @param dbpath the path to the KIR.dat file
#' @return A text representation of the KIR.dat file
readKirDat <- function(dbpath) {
  dbfile <- normalizePath(file.path(dbpath, "KIR.dat"), mustWork = TRUE)
  kirDat <- readLines(file(dbfile))
  unlink(dbpath, force = TRUE)
  kirDat
}

#' Get clostest complete neighbour allele
#'
#' @importFrom Biostrings getSeq
#' @importFrom Biostrings DNAStringSet
#' @import BSgenome

getClosest <- function(x){
  message("Infer the closest complete neighbours")
  # assertthat::assert_that(is("ipdDF", x))
  cmpl <- x$annot[x$annot$complete,]$GID
  ncmpl <- x$annot[!x$annot$complete,]$GID
  ann_ncmpl <- x$features[x$features$GID %in% ncmpl,]

  seqs <- DNAStringSet(x$seq$seq)
  names(seqs) <- x$seq$GID

  ## We need to know which partial genes are there
  featureGroups <- .getFeatureGroups(ann_ncmpl)
  message(sprintf("Found %s distinct feature compositions",
                  length(unique(featureGroups$feat))))

  ## Iterate over all possible partials and calc the corresponding dists
  allNeighbours <- lapply(unique(featureGroups$feat), function(featurePattern) {
    # featurePattern <- unique(featureGroups$feat)[1]
    # loi
    message(sprintf("Process feature composition %s", featurePattern))
    partialSeqs <- .getPartialSeqs(x$features, cmpl, featurePattern, seqs)
    neighbours <- lapply(unique(x$annot$locus), function(loi) {
      message(sprintf("Get closest neighbours for Locus %s", loi))
      .getDistMatrices(x$annot,
                       loi,
                       featureGroups,
                       featurePattern,
                       partialSeqs,
                       seqs)
    })
    do.call(rbind, neighbours)
  })
  do.call(rbind, allNeighbours)
}


## Get the distance matrix for each locus and report the nearest neighbour
.getDistMatrices <- function(annot, loi, featureGroups, featurePattern,
                             partialSeqs, seqs){
  allele_locus <- filter(annot, locus == loi)$GID
  aoi <- allele_locus[
    allele_locus %in%
      filter(featureGroups, feat == featurePattern)$GID]
  caoi <- names(partialSeqs)[names(partialSeqs) %in% allele_locus]
  allPartialSeqs <- c(partialSeqs[caoi],
                      seqs[names(seqs) %in% aoi])
  allPartialAln <- DECIPHER::AlignSeqs(allPartialSeqs,
                                       processors = getOption("ipd.ncores"))
  allPartialDm  <- DECIPHER::DistanceMatrix(
    allPartialAln, processors = getOption("ipd.ncores"))
  do.call(rbind, lapply(aoi, function(x) {
    data_frame(GID = x,
               closest_complete = names(which.min(allPartialDm[x,caoi])))
  }))
}

## Get the partial seqs composed of the desired features
.getPartialSeqs <- function(features, cmpl, featPattern, seqs){
  ranges <- GenomicRanges::GRangesList(BiocParallel::bplapply(cmpl, function(s)
    .extractRanges(s, featPattern, features)))
  partialSeqs <- getSeq(seqs, ranges)
  partialSeqs <- DNAStringSet(lapply(partialSeqs, unlist))
  names(partialSeqs) <- cmpl
  partialSeqs
}

.extractRanges <- function(allele, featPattern, features){
  featPattern <- strsplit(featPattern, ":")[[1]]
  alleleFeatures <- filter(features,
                           GID == allele,
                           feat_name %in% featPattern)
  GenomicRanges::GRanges(seqnames = alleleFeatures$GID,
                         ranges = IRanges::IRanges(start = alleleFeatures$start,
                                                   end = alleleFeatures$end,
                                                   names = alleleFeatures$GID))
}
## extract all possible and valid intron/exon combinations
.getFeatureGroups <- function(x){
  total_feats <- x %>%
    group_by(GID) %>%
    summarize(feat = paste(feat_name, collapse = ":"))
}

### Create HLA data without hlatools
###

#' Parse all HLA alleles for a locus from hla.xml
#'
#' @param doc \file{hla.xml} as an \code{xml_document} object.
#' @param locusname One of [\code{HLA-A, HLA-B, HLA-C, HLA-DPA1, HLA-DPB1,
#' HLA-DQA1, HLA-DQB1, HLA-DRB}]
#' @param ncores The number of compute cores to use.
#'
#' @return A \code{\linkS4class{HLAAllele}} object.
#'
parseHlaAlleles <- function(doc, locusnames = NULL,
                            ncores = getOption("ipd.ncores")) {
  if (is.null(locusnames))
    locusnames <- .VALID_HLA_LOCI()
  locusnames <- sapply(locusnames, .matchHlaLocus)
  if (is.null(ncores))
    ncores <- parallel::detectCores()/2
  assertthat::is.count(ncores)

  ns <- xml2::xml_ns(doc)
  xpaths1 <- sapply(locusnames, function(locusname)
    paste0("/d1:alleles/d1:allele/d1:locus[@locusname='",
           locusname,
           "']/parent::node()"))
  nodess1 <- lapply(xpaths1, function(xpath1) xml2::xml_find_all(doc,
                                                                 xpath1, ns))
  xpath2 <- paste0(".//d1:releaseversions[not(starts-with(@releasestatus,
                   'Allele Deleted'))]/parent::node()")
  nodes2 <- lapply(nodess1, function(nodes1) xml2::xml_find_all(nodes1,
                                                                xpath2, ns))

  releaseVersion <- xml2::xml_find_chr(
    nodes2[[1]][1], "string(./d1:releaseversions/@currentrelease)", ns)

  seq <- do.call(rbind, lapply(nodes2, .parseSequence))
  annot <- do.call(rbind,
                  lapply(1:length(nodes2),
                         function(lc)
                           .parseAnnotations(nodes2[[lc]],
                                             names(nodes2[lc]))))
  annot <- bind_cols(GID = annot$allele_name, annot)
  features <- do.call(rbind,
                      lapply(nodes2, function(x)
                        .parseFeatures(x, ncores = ncores)))

  list(
    seq      = seq,
    annot    = annot,
    features = features,
    version  = releaseVersion)
}

############# Helper to extract the data ###########
##### Extract sequence #####
.parseSequence = function(nodes) {
  ns     <- xml2::xml_ns(nodes)
  xpath  <- './d1:sequence/d1:nucsequence'
  data_frame(
    GID = xml2::xml_attr(nodes, "name"),
    seq     = xml2::xml_text(xml2::xml_find_all(nodes, xpath, ns)),
  )
}

## Extract annotations
.parseAnnotations = function(nodes, locus) {
  ns      <- xml2::xml_ns(nodes)
  res <- data_frame(
    ##
    ## Allele designation
    ##
    locus         = locus,
    allele_name   = xml2::xml_attr(nodes, "name"),
    allele_id     = xml2::xml_attr(nodes, "id"),
    g_group       = xml2::xml_find_chr(
      nodes, "string(./d1:hla_g_group/@status)", ns),
    p_group       = xml2::xml_find_chr(
      nodes, "string(./d1:hla_p_group/@status)", ns),
    date_assigned = xml2::xml_attr(nodes, "dateassigned"),
    ##
    ## Release
    ##
    first_released = xml2::xml_find_chr(
      nodes, "string(./d1:releaseversions/@firstreleased)", ns),
    last_updated   = xml2::xml_find_chr(
      nodes, "string(./d1:releaseversions/@lastupdated)", ns),
    release_status = xml2::xml_find_chr(
      nodes, "string(./d1:releaseversions/@releasestatus)", ns),
    confirmed      = xml2::xml_find_lgl(
      nodes, "string(./d1:releaseversions/@confirmed)=\"Confirmed\"", ns),
    ##
    ## CWD status and Completeness (we consider as complete alleles for which
    ## both UTRs are present)
    ##
    cwd_status    = xml2::xml_find_chr(
      nodes, "string(./d1:cwd_catalogue/@cwd_status)", ns),
    complete      = xml2::xml_find_lgl(
      nodes, "count(./d1:sequence/d1:feature[@featuretype=\"UTR\"])=2", ns),
    ##
  )
  res
}

##### Extract gene features ######
.parseFeatures = function(nodes, ncores) {
  doParallel::registerDoParallel(cores = ncores)
  ns       <- xml2::xml_ns(nodes)
  nodeset  <- xml2::xml_find_all(nodes, "./d1:sequence", ns)
  xpath    <- "./d1:feature[not(@featuretype=\"Protein\")]"
  seqnames <- xml2::xml_attr(nodes, "name")
  res <- foreach(node = nodeset, seqname = seqnames, .combine = rbind) %dopar% {
    data_frame(
      GID = seqname,
      start    = as.integer(xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/d1:SequenceCoordinates/@start"), ns))),
      end      = as.integer(xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/d1:SequenceCoordinates/@end"), ns))),
      feat_name = xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/@name"), ns)),
      id       = xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/@id"), ns)),
      feat_order    = as.integer(xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/@order"), ns))),
      type     = xml2::xml_text(xml2::xml_find_all(
        node, paste0(xpath, "/@featuretype"), ns)),
      frame    = vapply(xml2::xml_find_all(node, xpath, ns), function(node) {
        frame <- xml2::xml_find_chr(
          node, "string(./d1:cDNACoordinates/@readingframe)", ns)
        as.integer(ifelse(frame == "", 0, frame))
      }, FUN.VALUE = 0L)
      )
  }
  res
}





#' Parse all KIR alleles for a locus from KIR.dat
#'
#' @param doc \file{KIR.dat} as an line vector object.
#' @return A list of R6 class EmblEntry objects.
#'
readEmblEntries <- function(dat){
  iLine <- itertools::ihasNext(dat)
  emblEntryList <- c()
  idx <- 0
  while (TRUE){
    if (!itertools::hasNext(iLine))
      return(emblEntryList)
    line <- iterators::nextElem(iLine)
    idx <- idx+1
    line
    if (startsWith(line, "ID")){
      line = strsplit(strsplit(line, ";")[[1]][1], "\\s+")[[1]][2]
      eEntry <- EmblEntry(line)
      seqHeader <- ""
      deleted <- FALSE
    } else if (startsWith(line, "DT")){
      line <- strsplit(line, "\\s+")[[1]]
      if (line[5] == "Created,"){
        eEntry$setDateAssigned(line[2])
      } else if (line[6] == "Updated,"){
        eEntry$setLastUpdated(line[2])
      } else if (line[5] == "Current"){
        eEntry$setReleaseStatus(line[2])
      }
    } else if (startsWith(line, "FT")){
      line <- strsplit(line, "\\s+")[[1]]
      if (startsWith(line[2], "/allele")){
        features <- dplyr::data_frame(feat=character(),
                                      start=integer(),
                                      end=integer(),
                                      count=integer())
        allele <- strsplit(line[2], '\"')[[1]][2]
        eEntry$setAlleleName(allele)
      } else if (line[2] %in% c("UTR", "intron", "exon")){
        positions <- strsplit(line[3], "\\.\\.")[[1]]
        feature <- list(feat = line[2],
                        start = as.integer(positions[1]),
                        end = as.integer(positions[2]))
        if (line[2] == "UTR"){
          feature["count"] <- ifelse(NROW(features) == 0, 1L,2L)
          features <- dplyr::bind_rows(features, feature)
        }
      } else if (startsWith(line[2], "/number")){
        count = strsplit(line[2], '\"')[[1]][2]
        feature["count"] <- as.integer(gsub("/","", count))
        features <- dplyr::bind_rows(features, feature)
      }
    } else if (startsWith(line, "SQ")){
      if (is.null(eEntry$getAlleleName()))
        deleted <- TRUE
      eEntry$setFeat(features)

      seq
      seqHeader = eEntry$getAlleleName()
      seq <- Biostrings::DNAString()
    } else if (line == "//"){
      seq <- Biostrings::DNAStringSet(seq)
      names(seq) <- seqHeader
      eEntry$setSeq(seq)
      if (!is.null(eEntry$getAlleleName()))
        emblEntryList <- c(emblEntryList, eEntry)
    } else if (!seqHeader == "" && !deleted){
      line <- strsplit(line, "\\s+")[[1]]
      seqP <- Biostrings::DNAString(paste0(line[1:(length(line)-1)],
                                           collapse = ""))
      seq <- c(seq, seqP)
    }
    TRUE
  }
  emblEntryList
}

.emblEntryListToDataFrame <- function(embl){
  allele <- embl$getAlleleName()
  seq <- data_frame(GID = allele %||% NA,
                    seq = as.character(embl$getSeq()))
  features <- embl$getFeat()
  features$feat <- simpleCap(features$feat)
  features <- data_frame(GID        = allele %||% NA,
                              start      = features$start,
                              end        = features$end,
                              feat_name  = features$feat_name,
                              id         = "NA",
                              feat_order = features$count,
                              type       = features$feat,
                              frame      = 0
                              )
  annot <- data_frame(GID = allele %||% NA,
                      locus = embl$getLocus(),
                      allele_name = allele %||% NA ,
                      allele_id = embl$getId(),
                      g_group = "NA",
                      p_group = "NA",
                      date_assigned = embl$getDateAssigned() %||% "NA",
                      first_released = embl$getFirstReleased() %||% "NA",
                      last_updated = embl$getLastUpdated(),
                      release_status = embl$getReleaseStatus(),
                      confirmed = embl$getConfirmed() %||% NA,
                      cwd_status = "NA",
                      complete = embl$getComplete())

 list(seq = seq, features = features, annot = annot)
}


# Class: EmblEntry ----------------------------------------------------------

#' Constructor for EmblEntry objects.
#'
#' @param allele_id the id of a kir allele
#' @return A EmblEntry object

EmblEntry <- function(allele_id) {
  EmblEntry_$new(allele_id)
}

#' Class \code{"EmblEntry"}
#'
#' @docType class
#' @keywords data internal
#' @return Object of \code{\link{R6Class}} representing an Embl Entry.
#' @section Methods:
#' \describe{
#'   \item{\code{x$new(locusname, db_version, ncores = parallel::detectCores(),
#'     with_dist = FALSE)}}{Create an object of this class.}
#'   \item{\code{x$get_db_version()}}{Get the IMGT/HLA database version.}
#'   \item{\code{x$get_locusname()}}{Get the name of the locus.}
#'   \item{\code{x$get_alleles(allele)}}{Get alleles.}
#'   \item{\code{x$get_closest_complete_neighbor(allele)}}{Get the complete
#'     allele that is closest at exon 2 to the query allele.}
#'   \item{\code{x$get_reference_sequence(allele)}}{Get the (imputed) reference
#'     sequence for allele.}
#' }

EmblEntry_ <- R6::R6Class(
  classname = "EmblEntry",
  public = list(
    initialize = function(allele_id) {
      private$id <- allele_id
    },
    print = function() {
      fmt0 <- "EmblEntry <%s>;\nLocus: <%s>\nAllele: <%s>\nComplete: <%s>"
      cat(sprintf(fmt0,
                  self$getId(),
                  self$getLocus(),
                  self$getAlleleName(),
                  self$getComplete()))
      invisible(self)
    },
    ## getters and setters
    ## get Allele name
    getAlleleName = function() {
      private$name
    },
    ## set Allele name
    setAlleleName = function(allele_name){
      assertthat::is.string(allele_name)
      private$name = allele_name
    },
    ## set sequence
    setSeq = function(seq){
      assertthat::assert_that(is(object = seq, class2 = "DNAStringSet"))
      private$seq = seq
      return(self)
    },
    ## get Sequence
    getSeq = function(){
      private$seq
    },
    ## set features
    setFeat = function(features){
      private$feat = features
      return(self)
    },
    ## get features
    getFeat = function(){
      .getFeatures(private$feat)
    },
    ## get Locus
    getLocus = function(){
      strsplit(private$name, "\\*")[[1]][1]
    },
    ## set Id
    setId = function(id){
      assertthat::is.string(id)
      private$id = id
      return(self)
    },
    ## get allele id
    getId = function(){
      private$id
    },
    ## set date assigned
    setDateAssigned = function(date){
      ## Todo add test for object
      private$date_assigned = date
      return(self)
    },
    ## get date assigned
    getDateAssigned = function(){
      private$date_assigned
    },
    ## set first_released
    setFirstReleased = function(date){
      ## Todo add test for object
      private$first_released = date
      return(self)
    },
    ## get first released
    getFirstReleased = function(){
      private$first_released
    },
    ## set release status
    setReleaseStatus = function(date){
      ## Todo add test for object
      private$release_status = date
      return(self)
    },
    ## get release status
    getReleaseStatus = function(){
      private$release_status
    },
    ## set last updated
    setLastUpdated = function(date){
      ## Todo add test for object
      private$last_updated = date
      return(self)
    },
    ## get last updated
    getLastUpdated= function(){
      private$last_updated
    },
    ## get confirmed
    setConfirmed = function(confirmed){
      assertthat::assert_that(is.logical(confirmed))
      private$confirmed = confirmed
      return(self)
    },
    ## get confirmed
    getConfirmed = function(){
      private$confirmed
    },
    ## get complete
    getComplete = function(){
      sum(self$getFeat()$feat == "UTR") == 2
    }
  ),
  private = list(
    name = NULL, # [character]; allele name
    seq  = NULL, # [DNAStringSet]: allele sequence
    feat = NULL, # [GRanges]: feature ranges
    lcn  = NULL, # [character]; locus name
    id   = NULL, # [character]: allele id
    date_assigned = NULL,
    first_released = NULL,
    last_updated = NULL,
    release_status = NULL,
    confirmed = NULL,
    cwd_status = NULL
  )
)
##### Helper #####
#####

.getFeatures <- function(feats){
  feats %>%
    dplyr::mutate(feat_name = ifelse(
      feat == "UTR",
      ifelse(count == 1,
             "3' UTR",
             "5' UTR"),
      paste(paste0(toupper(substring(feat, 1, 1)),
                   substring(feat, 2, nchar(feat))),
            count, sep = " " )))
}

#### Utils ####

.VALID_HLA_LOCI <- function(){
  c("HLA-A", "HLA-B", "HLA-C", "HLA-DPB1", "HLA-DQB1", "HLA-DRB1")
}
.VALID_KIR_LOCI <- function(){
    c("KIR2DL1",  "KIR2DL2",  "KIR2DL3",  "KIR2DL4", "KIR2DL5A", "KIR2DL5B",
      "KIR2DP1",  "KIR2DS1", "KIR2DS2",  "KIR2DS3",  "KIR2DS4",  "KIR2DS5",
      "KIR3DL1",  "KIR3DL2",  "KIR3DL3",  "KIR3DP1", "KIR3DS1")
  }

.expandHlaAllele <- function(x, locus = NULL) {
  if (is.null(locus)) {
    ifelse(!grepl("^HLA-\\S+", x), paste0("HLA-", x), x)
  } else {
    locus <- sub("HLA-", "", toupper(locus))
    pattern1 <- paste0("^HLA-", locus, "[*]\\d\\d\\d?:.+$")
    pattern2 <- paste0("^", locus, "[*]\\d\\d\\d?:.+$")
    pattern3 <- "^\\d\\d\\d?:.+$"
    ifelse(grepl(pattern1, x),
           x,
           ifelse(grepl(pattern2, x),
                  paste0("HLA-", x),
                  ifelse(grepl(pattern3, x),
                         paste0("HLA-", locus, "*", x), x)))
  }
}

.matchHlaLocus <- function(locus) {
  locus <- .expandHlaAllele(locus)
  match.arg(locus, .VALID_HLA_LOCI())
}

.expandKirAllele <- function(x, locus = NULL) {
  if (is.null(locus)) {
    ifelse(!grepl("^HLA-\\S+", x), paste0("HLA-", x), x)
  } else {
    locus <- sub("HLA-", "", toupper(locus))
    pattern1 <- paste0("^HLA-", locus, "[*]\\d\\d\\d?:.+$")
    pattern2 <- paste0("^", locus, "[*]\\d\\d\\d?:.+$")
    pattern3 <- "^\\d\\d\\d?:.+$"
    ifelse(grepl(pattern1, x),
           x,
           ifelse(grepl(pattern2, x),
                  paste0("HLA-", x),
                  ifelse(grepl(pattern3, x),
                         paste0("HLA-", locus, "*", x), x)))
  }
}

.matchKirLocus <- function(locus) {
  locus <- .expandKirAllele(locus)
  match.arg(locus, .VALID_KIR_LOCI())
}
simpleCap <- function(x) {
  sapply(x, function(e){
    s <- strsplit(e, " ")[[1]]
    paste(toupper(substring(s, 1,1)), substring(s, 2),
          sep="", collapse=" ")
  })
}

`%||%` <- function(a, b) {
  if (length(a) == 0) b else a
}

#' gather the Hla data

#'  fetch every information and create the dataframes
#' Get the HLA info
#' @return A data.frame with all information to be stored in the database
getHlaDf <- function(){
  hlaDb <- fetchIPD(family = "HLA")
  hladoc <- readHlaXml(hlaDb)
  hla <- parseHlaAlleles(hladoc)
  closestComplete <- getClosest(hla)
  hla$annot <- dplyr::left_join(hla$annot, closestComplete, by="GID")
  hla
}

#' gather the KIR data
#'
#' fetch every information and create the dataframes
#' Get the HLA info
#' @return A data.frame with all information to be stored in the database
getKirDf <- function() {
  kirDb <- fetchIPD(family = "KIR")
  kirdoc <- readKirDat(kirDb)
  pattern <- "^.*(\\d+\\.\\d+\\.\\d+).*"
  version <- sub(pattern, "\\1",
                 kirdoc[grep("CC   IPD-KIR Release Version", kirdoc)[1]])
  kirObj <- readEmblEntries(kirdoc)
  kirFullDFList <- lapply(kirObj, .emblEntryListToDataFrame)
  kirDf <- list()
  kirDf$seq <- do.call(rbind, lapply(kirFullDFList, function(x) x$seq))
  kirDf$features <- do.call(rbind,
                            lapply(kirFullDFList, function(x) x$features))
  kirDf$annot <- do.call(rbind, lapply(kirFullDFList, function(x) x$annot))
  closestComplete <- getClosest(kirDf)
  kirDf$annot <- dplyr::left_join(kirDf$annot, closestComplete, by="GID")
  kirDf$version <- version
  kirDf
}

#' Prepare the final database
#'
#' Some restructuring of the data to fit into the SQLite database.
#'
#' @param df the dataframe from either getKirDf() or getHlaDf().
#' @return a list containing the final data.frame and version information.
#'
prepareDb <- function(df){
  ipd <- list()
  ipd$annot <- df$annot %>%
    mutate(closest_complete = ifelse(is.na(closest_complete), allele_name,
                                     closest_complete)) %>%
    mutate(g_group = ifelse(is.na(g_group), "none", g_group)) %>%
    mutate(p_group = ifelse(is.na(p_group), "none", p_group)) %>%
    mutate(confirmed = ifelse(is.na(confirmed), "NA", confirmed)) %>%
    mutate(first_released = ifelse(is.na(first_released), "NA",
                                   first_released)) %>%
    mutate(date_assigned = ifelse(is.na(date_assigned), "NA",
                                  date_assigned)) %>%
    mutate(cwd_status = ifelse(is.na(cwd_status), "NA", cwd_status))
  ipd$features <- df$features
  ipd$seq <- df$seq
  list(ipd = ipd, version = df$version)
}

#' Get the Kir data ready
#'
#' @return a list containing the final data.frame and version information.
getKir <- function() {
  df <- getKirDf()
  prepareDb(df)
}

#' Get the Hla Data ready
#'
#' @return a list containing the final data.frame and version information.
getHla <- function() {
  df <- getHlaDf()
  prepareDb(df)
}

## Run the functions and get the actual data
hla <- getHla()
kir <- getKir()

hlaMetadata <- data.frame(
  Title = "Allele data from the IPD IMGT/HLA database version 3.32.0",
  Description = paste0("Data for all alleles of selected HLA loci (HLA-A, -B, 
  -C, -DPB1, -DQB1 and -DRB1). The allele annotation, sequence, gene structure
  and the (sequence-based) closest allele in full-length is stored. 
  Reference:
  Robinson J, Maccari G, Marsh SGE, Walter L, Blokhuis J, Bimber B, Parham P, 
  De Groot NG, Bontrop RE, Guethlein LA, and Hammond JA
  KIR Nomenclature in non-human species Immunogenetics (2018), in preparation"),
  BiocVersion = "3.8",
  Genome = "no genome data",
  SourceType = "Zip",
  SourceUrl = "https://github.com/ANHIG/IMGTHLA/blob/Latest/xml/hla.xml.zip",
  SourceVersion = "3.32.0",
  Species = "Homo sapiens",
  TaxonomyId = 9606,
  Coordinate_1_based = TRUE,
  DataProvider = "EMBL-EBI",
  Maintainer = "Steffen Klasberg <klasberg@dkms-lab.de>",
  RDataClass = "data.frame, DNAStringSet, GRanges",
  DispatchClass = "SQLiteFile",
  RDataPath = "ipdDb/ipdHLA_3.32.0.sqlite",
  Tags = "ipd:hla:IMGT/HLA:alleles"
)
hlaMetadata <- rbind(hlaMetadata, data.frame(
  Title = "Allele data from the IPD IMGT/HLA database version 3.33.0",
  Description = paste0("Data for all alleles of selected HLA loci (HLA-A, -B, 
  -C, -DPB1, -DQB1 and -DRB1). The allele annotation, sequence, gene structure
  and the (sequence-based) closest allele in full-length is stored. 
  Reference:
  Robinson J, Maccari G, Marsh SGE, Walter L, Blokhuis J, Bimber B, Parham P, 
  De Groot NG, Bontrop RE, Guethlein LA, and Hammond JA
  KIR Nomenclature in non-human species Immunogenetics (2018), in preparation"),
  BiocVersion = "3.8",
  Genome = "no genome data",
  SourceType = "Zip",
  SourceUrl = "https://github.com/ANHIG/IMGTHLA/blob/Latest/xml/hla.xml.zip",
  SourceVersion = "3.33.0",
  Species = "Homo sapiens",
  TaxonomyId = 9606,
  Coordinate_1_based = TRUE,
  DataProvider = "EMBL-EBI",
  Maintainer = "Steffen Klasberg <klasberg@dkms-lab.de>",
  RDataClass = "data.frame, DNAStringSet, GRanges",
  DispatchClass = "SQLiteFile",
  RDataPath = "ipdDb/ipdHLA_3.33.0.sqlite",
  Tags = "ipd:hla:IMGT/HLA:alleles"
))
kirMetadata <- data.frame(
  Title = "Allele data from the IPD KIR database version 2.7.1",
  Description = paste0("Data for the alleles of all KIR loci in the database. 
  The allele annotation, sequence, gene structure and the (sequence-based) 
  closest allele in full-length is stored. 
  Reference:
  Robinson J, Maccari G, Marsh SGE, Walter L, Blokhuis J, Bimber B, Parham P, 
  De Groot NG, Bontrop RE, Guethlein LA, and Hammond JA
  KIR Nomenclature in non-human species Immunogenetics (2018), in preparation"),
  BiocVersion = "3.8",
  Genome = "no genome data",
  SourceType = "Zip",
  SourceUrl = "https://github.com/ANHIG/IPDKIR/blob/Latest/KIR.dat",
  SourceVersion = "2.7.1",
  Species = "Homo sapiens",
  TaxonomyId = 9606,
  Coordinate_1_based = TRUE,
  DataProvider = "EMBL-EBI",
  Maintainer = "Steffen Klasberg <klasberg@dkms-lab.de>",
  RDataClass = "data.frame, DNAStringSet, GRanges",
  DispatchClass = "SQLiteFile",
  RDataPath = "ipdDb/ipdKIR_2.7.1.sqlite",
  Tags = "ipd:kir:alleles"
)

write.csv(rbind(hlaMetadata, kirMetadata), "inst/extdata/ipd_metadata.csv")

## Get the reference sequences as Biostrings::DNAStringSets
#
.getReference <- function(x, alleles) {
  cols <- c("seq")
  keytype <- "allele_name"
  res <- suppressMessages(
    select(x, keys = alleles, columns = cols, keytype = keytype))
  seqs <- DNAStringSet(res$seq)
  names(seqs) <- res$allele_name
  seqs
}

## Get gene structure as GenomicRanges::GRanges objects
##
.getStructure <- function(x, alleles) {
  cols <- c("start", "end", "type", "feat_name",
            "complete", "frame")
  keytype <- "allele_name"
  res <- suppressMessages(
    select(x, keys = alleles, columns = cols, keytype =keytype ))
  GRanges(seqnames = res$allele_name,
          ranges = IRanges(start = as.integer(res$start),
                           end = as.integer(res$end),
                           names = res$allele_name),
          strand = "+",
          type = res$type,
          name = res$feat_name,
          complete = as.logical(as.integer(res$complete)),
          frame  = as.integer(res$frame))
}

## Get the clostest complete reference
#
.getClosestComplete <- function(x, allele, locus = NULL) {
  assert_that(is.string(allele))
  if (!is.null(locus))
    allele <- .matchAllele(x, allele, locus)
  cols <- c("complete", "closest_complete")
  keytype <- "allele_name"
  res <- suppressMessages(
    select(x, keys = allele, columns = cols, keytype = keytype))
  if (NROW(res) == 0) {
    stop("Allele not found. Try again with adding the locus to the call")
  }
  closestComplete <- ifelse(length(allele) == 1,
                            res$closest_complete,
                            res[res$complete == 1, ][1, ]$closest_complete)

  getReference(x, closestComplete)
}

## Get all alleles of a locus
##
.getAlleles <- function(x, locus) {
  keytypes(x)
  keys(x, keytype = "TYPE")
  cols <- c("allele_name")
  res <- suppressMessages(
    select(x, keys = locus, keytype = "LOCUS", columns = cols))
  res$allele_name
}

## Match partial alleles
##
.matchAllele <- function(x, allele, locus) {
  assert_that(locus %in% getLoci(x))
  assert_that(is.character(allele))
  validAlleles <- getAlleles(x, locus)
  if (allele %in% validAlleles) {
    message(sprintf("Found allele %s", allele ))
    return(allele)
  } else if (any(matchingAlleles <- startsWith(validAlleles, allele))) {
    allele <- validAlleles[matchingAlleles]
    message(sprintf("Found allele %s by extension", allele))
    return(allele)
  } else if (any(matchingAlleles <- grepl(allele, validAlleles))) {
    allele <- validAlleles[matchingAlleles]
    message(sprintf("Found allele %s by regex matching", allele))
    return(allele)
  } else {
    message(
      sprintf("Could not find allele %s. Match to all of the locus", allele))
    return(validAlleles)
  }
}

## Get all loci
##
.getLoci <- function(x) {
  assert_that(is(x, "IpdDb"))
  unique(keys(x, keytype = "locus"))
}

## Create the select() interface
##  Columns and keytypes method ##
.getDataTables <- function(con){
  tables <- RSQLite::dbListTables(con)
  tables <- tables[!tables %in% c("metadata","map_metadata","map_counts")]
  setNames(tables, tables)
}

.getLCcolnames <- function(x){
  con <- dbconn(x)
  tables <- .getDataTables(con)
  cols <- unique(unlist(lapply(tables, FUN=RSQLite::dbListFields, con=con)))
  cols <- cols[!cols %in% "_id"]
  cols
}

.cols <- function(x){
  list <- .getLCcolnames(x)
  ## dont use the primary keytype
  list <- list[!list == "GID"]
  toupper(list)
}


## Keys method ##
.deriveTableNameFromField <- function(field, x){
  con <- dbconn(x)
  tables <- .getDataTables(con)
  colTabs <- lapply(tables, FUN=RSQLite::dbListFields, con=con)
  m <- unlist2(lapply(colTabs, match, field))
  tab <- names(m)[!is.na(m)]
  if(length(tab) > 1){stop("Two fields in the source DB have the same name.")}
  if(length(tab) == 0){stop("Did not find a field in the source DB.")}
  tab
}

.keys <- function(x, keytype){
  keytype <- tolower(keytype)
  tab <- .deriveTableNameFromField(field=keytype, x)
  ## So now we know table name (tab) and field (keytype)
  sql <- paste("SELECT",keytype,"FROM",tab)
  res <- RSQLite::dbGetQuery(dbconn(x), sql)
  as.character(res[!is.na(res)])
}


.testForValidKeys <- function(x, keys, keytype){
  if (!is.character(keys)){
    stop("'keys' must be a character vector")
  }
  if (length(keys) == 0L) {
    return()
  }
  ktKeys <- keys(x, keytype)

  if(!(any(ktKeys %in% keys))){
    msg <- paste0("None of the keys entered are valid keys for '",keytype,
                  "'. Use the keys method to see a listing of valid arguments.")
    stop(msg) ## later when things are better, demote this to a warning()
    return(FALSE)
  }
  return(TRUE)
}

## select methods
##
.select <- function(x, keys=NULL, columns=NULL, keytype){
  keytype <- tolower(keytype)
  columns <- tolower(columns)
  .testForValidKeys(x, keys, keytype)
  ## 1st pool all the fields we need to extract
  fields <- unique(c(columns, keytype))
  ## Then get the tables to go with each one.
  tabs <- vapply(fields, .deriveTableNameFromField, x=x,
                 FUN.VALUE = character(1))
  ## make fully qualified fields of these tabs (the ones we want to extract)
  f.fields <- paste(tabs, fields, sep=".")

  ## Make non-redundant list of tables to visit
  nrTabs <- unique(tabs)
  ## Now join to each table
  for(i in seq_along(nrTabs)){
    if(i==1){
      sql <- paste("SELECT ",paste(f.fields, collapse=","),
                   " FROM",tabs[1])
    }else{
      ## IF we see c.probes in nrTabs[i], it means we have to
      ## use gene_id instead.
      if("c.probes" %in% nrTabs[i]){
        sql <- c(sql, paste("LEFT JOIN ",nrTabs[i],"USING (GID)"))
      }else{
        sql <- c(sql, paste("LEFT JOIN ",nrTabs[i],"USING (_id)"))
      }
    }
  }
  sql <- paste(sql, collapse=" ")
  ## add the where clause
  fullKeytype <- tabs[names(tabs)==keytype]
  fullKeytype <- paste(fullKeytype, names(fullKeytype), sep=".")
  strKeys <- paste0('"',keys,'"',collapse=",")
  where <- paste("WHERE ",fullKeytype,"IN (",strKeys,")" )
  sql <- paste(sql, where)
  ## then call that
  res <- RSQLite::dbGetQuery(dbconn(x), sql)
  ## cleanup and re-organize
  resort_base(res, keys, jointype=keytype, fields)
}

#' Get allele information from IPD
#' 
#' This package holds the IPD IMGT/HLA and IPD KIR database.
#' All alleles are accessible using the select, columns, keys and keytypes
#' methods of the AnnoatationDbi package of bioconductor.
#' 
#' Included data are:
#'   
#' Allele names
#' 
#' p-groups
#' 
#' g-groups
#' 
#' cwd_status
#' 
#' completeness status
#' 
#' gene structure
#' 
#' reference sequences
#' 
#' closest full-length allele
#' 
#' @docType package
#' @name ipdDbPackage
#' @importFrom Biostrings DNAStringSet
#' @importFrom GenomicRanges GRanges
#' @importFrom IRanges IRanges
#' @importFrom stats setNames
#' @import methods
#' @import AnnotationDbi
#' @import RSQLite
#' @import DBI
#' @import assertthat
#' @import AnnotationHub

NULL

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{methods}
\name{getReference,IpdDb-method}
\alias{getReference,IpdDb-method}
\alias{getReference}
\title{Get reference sequences}
\usage{
getReference(x, allele)
}
\arguments{
\item{x}{The database connection; an \code{\link{IpdDb}} object.}

\item{allele}{The alleles of interest as a character vector.}
}
\value{
A \code{\link[Biostrings:XStringSet-class]{Biostrings:DNAStringSet}}
object with all references.
}
\description{
Get the reference sequences for alleles.
}
\examples{
## Load the database 
hla <- loadHlaData()
## Get the loci
loci <- getLoci(hla)
## Get alleles of a locus
alleles <- getAlleles(hla, loci[1])
allelesOfInterest <- alleles[1:10]
## Get the sequences
seqs <- getReference(hla, allelesOfInterest)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{methods}
\name{getAlleles,IpdDb-method}
\alias{getAlleles,IpdDb-method}
\alias{getAlleles}
\title{Get alleles}
\usage{
getAlleles(x, locus)
}
\arguments{
\item{x}{The database connection; an \code{\link{IpdDb}} object.}

\item{locus}{A single locus as a string.}
}
\value{
A character vector with all alleles of the give locus.
}
\description{
Get all alleles of a given locus.
}
\examples{
## Load the database 
hla <- loadHlaData()
## Get the loci
loci <- getLoci(hla)
## Get alleles of a locus
alleles <- getAlleles(hla, loci[1])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{methods}
\name{getLoci,IpdDb-method}
\alias{getLoci,IpdDb-method}
\alias{getLoci}
\title{Get loci}
\usage{
getLoci(x)
}
\arguments{
\item{x}{The database connection; an \code{\link{IpdDb}} object.}
}
\value{
A vector of available loci in the database.
}
\description{
Get all available loci of the KIR or HLA database
}
\examples{
## Load the database 
hla <- loadHlaData()
## Get the loci
loci <- getLoci(hla)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{class}
\name{IpdDb-class}
\alias{IpdDb-class}
\alias{.IpdDb}
\alias{IpdDb}
\alias{columns,IpdDb-method}
\alias{columns}
\alias{keytypes,IpdDb-method}
\alias{keytypes}
\alias{keys,IpdDb-method}
\alias{keys}
\alias{select,IpdDb-method}
\alias{select}
\title{The database class for storing allele data from IPD.}
\usage{
columns(x)

keytypes(x)

keys(x, keytype, ...)

select(x, keys, columns, keytype, ...)
}
\arguments{
\item{x}{the IpdDb object}

\item{keytype}{The keytype for which the keys are retrieved}

\item{...}{Additional arguments. Not used now.}

\item{keys}{The keys for which columns should be selected by select()}

\item{columns}{The columns to retrieve by select}
}
\value{
character vector (keys, columns, keytypes) or a data.frame (select).
}
\description{
This class extends the 
\code{\link[AnnotationDbi:AnnDbObj-class]{AnnotationDbi::AnnDbObj-class}}
object by higher level methods for sequence and annotation retrieval.
blubb
}
\section{Slots}{

\describe{
\item{\code{getDbVersion()}}{Get the version of the original ipd database}

\item{\code{getLoci()}}{get all loci from a database, see 
\code{\link{getLoci}}.}

\item{\code{getReference(alleles)}}{Get the reference sequence for alleles, see 
\code{\link{getReference}}.}

\item{\code{getStructure(alleles)}}{Get the structures of alleles, see 
\code{\link{getStructure}}.}

\item{\code{getClosestComplete(allele)}}{Get the closest full-length reference 
sequence of one allele, see \code{\link{getClosestComplete}}.}

\item{\code{getAlleles(locus)}}{Get all alleles of a locus, see 
\code{\link{getAlleles}}.}
}}


\examples{
## load the data 
hla <- loadHlaData()
## get all valid keytypes
kts <- keytypes(hla)
## get all valid columns
cols <- columns(hla)
## get the keys of one keytype
kt <- kts[1]
keys <- keys(hla, kt)
## Get data of the two first columns for the first 10 keys 
cols <- cols[1:10]
res <- select(hla, keys, cols, kt)
  
}
\seealso{
\code{\link[AnnotationDbi:AnnDbObj-class]{AnnotationDbi::AnnDbObj-class}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{methods}
\name{getClosestComplete,IpdDb-method}
\alias{getClosestComplete,IpdDb-method}
\alias{getClosestComplete}
\title{Get closest full-length sequence}
\usage{
getClosestComplete(x, allele, locus = NULL)
}
\arguments{
\item{x}{The database connection; an \code{\link{IpdDb}} object.}

\item{allele}{A single allele as a string.}

\item{locus}{optional parameter used if the allele identifier is not found.}
}
\value{
A \code{\link[Biostrings:XStringSet-class]{Biostrings:DNAStringSet}}
object with the sequence 
of the closest full-length allele.
}
\description{
Get the sequence of the closest allele which for which a full-length 
  sequence is available.
}
\examples{
## Load the database 
hla <- loadHlaData()
## Get the loci
loci <- getLoci(hla)
## Get alleles of a locus
alleles <- getAlleles(hla, loci[1])
alleleOfInterest <- alleles[1]
## Get the closest complete sequence
seqs <- getClosestComplete(hla, alleleOfInterest, loci[1])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ipdDbClass.R
\docType{methods}
\name{getStructure,IpdDb-method}
\alias{getStructure,IpdDb-method}
\alias{getStructure}
\title{Get gene structures}
\usage{
getStructure(x, allele)
}
\arguments{
\item{x}{The database connection; an \code{\link{IpdDb}} object.}

\item{allele}{The alleles of interest as a character vector.}
}
\value{
A \code{\link[GenomicRanges:GRanges-class]{GenomicRanges:GRanges}} 
object with all gene
  structures.
}
\description{
Get the gene structures for alleles.
}
\examples{
## Load the database 
hla <- loadHlaData()
## Get the loci
loci <- getLoci(hla)
## Get alleles of a locus
alleles <- getAlleles(hla, loci[1])
allelesOfInterest <- alleles[1:10]
## Get the structures
seqs <- getStructure(hla, allelesOfInterest)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadDb.R
\name{loadKirData}
\alias{loadKirData}
\title{Load the IPD KIR database}
\usage{
loadKirData(version = "Latest")
}
\arguments{
\item{version}{Either a valid version of the IPD KIR database or "Latest" to
fetch the latest version}
}
\value{
an \code{\link{IpdDb}} object containing the database.
}
\description{
Load the IPD KIR database
}
\examples{
## Load the KIR database
kir <- loadKirData()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IpdDb.R
\docType{package}
\name{ipdDbPackage}
\alias{ipdDbPackage}
\alias{ipdDbPackage-package}
\title{Get allele information from IPD}
\description{
This package holds the IPD IMGT/HLA and IPD KIR database.
All alleles are accessible using the select, columns, keys and keytypes
methods of the AnnoatationDbi package of bioconductor.
}
\details{
Included data are:
  
Allele names

p-groups

g-groups

cwd_status

completeness status

gene structure

reference sequences

closest full-length allele
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loadDb.R
\name{loadHlaData}
\alias{loadHlaData}
\title{Load the IPD IMGT/HLA database}
\usage{
loadHlaData(version = "Latest")
}
\arguments{
\item{version}{Either a valid version of the IPD IMGT/HLA database or 
"Latest" to fetch the latest version}
}
\value{
an \code{\link{IpdDb}} object containing the database.
}
\description{
Load the IPD IMGT/HLA database
}
\examples{
## Load the HLA database
hla <- loadHlaData()
}
